@@ -1,42 +1,16 @@
-///////////////////////////////////////////////////////////////////////////
-//
-// Copyright (c) 2012-2016 DreamWorks Animation LLC
-//
-// All rights reserved. This software is distributed under the
-// Mozilla Public License 2.0 ( http://www.mozilla.org/MPL/2.0/ )
-//
-// Redistributions of source code must retain the above copyright
-// and license notice and the following restrictions and disclaimer.
-//
-// *     Neither the name of DreamWorks Animation nor the names of
-// its contributors may be used to endorse or promote products derived
-// from this software without specific prior written permission.
-//
-// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY INDIRECT, INCIDENTAL,
-// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-// IN NO EVENT SHALL THE COPYRIGHT HOLDERS' AND CONTRIBUTORS' AGGREGATE
-// LIABILITY FOR ALL CLAIMS REGARDLESS OF THEIR BASIS EXCEED US$250.00.
-//
-///////////////////////////////////////////////////////////////////////////
+// Copyright Contributors to the OpenVDB Project
+// SPDX-License-Identifier: MPL-2.0
 
 #ifndef OPENVDB_MATH_MAT3_H_HAS_BEEN_INCLUDED
 #define OPENVDB_MATH_MAT3_H_HAS_BEEN_INCLUDED
 
-#include <iomanip>
-#include <assert.h>
-#include <math.h>
 #include <openvdb/Exceptions.h>
 #include "Vec3.h"
 #include "Mat.h"
+#include <algorithm> // for std::copy()
+#include <cassert>
+#include <cmath>
+#include <iomanip>
 
 
 namespace openvdb {
@@ -55,11 +29,14 @@ class Mat3: public Mat<3, T>
 {
 public:
     /// Data type held by the matrix.
-    typedef T         value_type;
-    typedef T         ValueType;
-    typedef Mat<3, T> MyBase;
+    using value_type = T;
+    using ValueType = T;
+    using MyBase = Mat<3, T>;
+
     /// Trivial constructor, the matrix is NOT initialized
-    Mat3() {}
+    /// @note destructor, copy constructor, assignment operator and
+    ///   move constructor are left to be defined by the compiler (default)
+    Mat3() = default;
 
     /// Constructor given the quaternion rotation, e.g.    Mat3f m(q);
     /// The quaternion is normalized and used to construct the matrix
@@ -78,15 +55,15 @@ public:
          Source d, Source e, Source f,
          Source g, Source h, Source i)
     {
-        MyBase::mm[0] = static_cast<ValueType>(a);
-        MyBase::mm[1] = static_cast<ValueType>(b);
-        MyBase::mm[2] = static_cast<ValueType>(c);
-        MyBase::mm[3] = static_cast<ValueType>(d);
-        MyBase::mm[4] = static_cast<ValueType>(e);
-        MyBase::mm[5] = static_cast<ValueType>(f);
-        MyBase::mm[6] = static_cast<ValueType>(g);
-        MyBase::mm[7] = static_cast<ValueType>(h);
-        MyBase::mm[8] = static_cast<ValueType>(i);
+        MyBase::mm[0] = static_cast<T>(a);
+        MyBase::mm[1] = static_cast<T>(b);
+        MyBase::mm[2] = static_cast<T>(c);
+        MyBase::mm[3] = static_cast<T>(d);
+        MyBase::mm[4] = static_cast<T>(e);
+        MyBase::mm[5] = static_cast<T>(f);
+        MyBase::mm[6] = static_cast<T>(g);
+        MyBase::mm[7] = static_cast<T>(h);
+        MyBase::mm[8] = static_cast<T>(i);
     } // constructor1Test
 
     /// Construct matrix from rows or columns vectors (defaults to rows
@@ -108,27 +85,17 @@ public:
     template<typename Source>
     Mat3(Source *a)
     {
-        MyBase::mm[0] = a[0];
-        MyBase::mm[1] = a[1];
-        MyBase::mm[2] = a[2];
-        MyBase::mm[3] = a[3];
-        MyBase::mm[4] = a[4];
-        MyBase::mm[5] = a[5];
-        MyBase::mm[6] = a[6];
-        MyBase::mm[7] = a[7];
-        MyBase::mm[8] = a[8];
+        MyBase::mm[0] = static_cast<T>(a[0]);
+        MyBase::mm[1] = static_cast<T>(a[1]);
+        MyBase::mm[2] = static_cast<T>(a[2]);
+        MyBase::mm[3] = static_cast<T>(a[3]);
+        MyBase::mm[4] = static_cast<T>(a[4]);
+        MyBase::mm[5] = static_cast<T>(a[5]);
+        MyBase::mm[6] = static_cast<T>(a[6]);
+        MyBase::mm[7] = static_cast<T>(a[7]);
+        MyBase::mm[8] = static_cast<T>(a[8]);
     } // constructor1Test
 
-    /// Copy constructor
-    Mat3(const Mat<3, T> &m)
-    {
-        for (int i=0; i<3; ++i) {
-            for (int j=0; j<3; ++j) {
-                MyBase::mm[i*3 + j] = m[i][j];
-            }
-        }
-    }
-
     /// Conversion constructor
     template<typename Source>
     explicit Mat3(const Mat3<Source> &m)
@@ -152,11 +119,21 @@ public:
 
     /// Predefined constant for identity matrix
     static const Mat3<T>& identity() {
+        static const Mat3<T> sIdentity = Mat3<T>(
+            1, 0, 0,
+            0, 1, 0,
+            0, 0, 1
+        );
         return sIdentity;
     }
 
     /// Predefined constant for zero matrix
     static const Mat3<T>& zero() {
+        static const Mat3<T> sZero = Mat3<T>(
+            0, 0, 0,
+            0, 0, 0,
+            0, 0, 0
+        );
         return sZero;
     }
 
@@ -194,20 +171,6 @@ public:
         return Vec3<T>((*this)(0,j), (*this)(1,j), (*this)(2,j));
     } // rowColumnTest
 
-    // NB: The following two methods were changed to
-    // work around a gccWS5 compiler issue related to strict
-    // aliasing (see FX-475).
-
-    //@{
-    /// Array style reference to ith row
-    /// e.g.    m[1][2] = 4;
-    T* operator[](int i) { return &(MyBase::mm[i*3]); }
-    const T* operator[](int i) const { return &(MyBase::mm[i*3]); }
-    //@}
-
-    T* asPointer() {return MyBase::mm;}
-    const T* asPointer() const {return MyBase::mm;}
-
     /// Alternative indexed reference to the elements
     /// Note that the indices are row first and column second.
     /// e.g.    m(0,0) = 1;
@@ -228,7 +191,7 @@ public:
         return MyBase::mm[3*i+j];
     } // trivial
 
-    /// Set the rows of "this" matrix to the vectors v1, v2, v3
+    /// Set the rows of this matrix to the vectors v1, v2, v3
     void setRows(const Vec3<T> &v1, const Vec3<T> &v2, const Vec3<T> &v3)
     {
         MyBase::mm[0] = v1[0];
@@ -241,8 +204,8 @@ public:
         MyBase::mm[7] = v3[1];
         MyBase::mm[8] = v3[2];
     } // setRows
-    
-    /// Set the columns of "this" matrix to the vectors v1, v2, v3
+
+    /// Set the columns of this matrix to the vectors v1, v2, v3
     void setColumns(const Vec3<T> &v1, const Vec3<T> &v2, const Vec3<T> &v3)
     {
         MyBase::mm[0] = v1[0];
@@ -256,12 +219,6 @@ public:
         MyBase::mm[8] = v3[2];
     } // setColumns
 
-    /// Set the rows of "this" matrix to the vectors v1, v2, v3
-    OPENVDB_DEPRECATED void setBasis(const Vec3<T> &v1, const Vec3<T> &v2, const Vec3<T> &v3)
-    {
-        this->setRows(v1, v2, v3);
-    }
-
     /// Set diagonal and symmetric triangular components
     void setSymmetric(const Vec3<T> &vdiag, const Vec3<T> &vtri)
     {
@@ -276,8 +233,7 @@ public:
         MyBase::mm[8] = vdiag[2];
     } // setSymmetricTest
 
-    /// Returns matrix with prescribed diagonal and symmetric triangular
-    /// components
+    /// Return a matrix with the prescribed diagonal and symmetric triangular components.
     static Mat3 symmetric(const Vec3<T> &vdiag, const Vec3<T> &vtri)
     {
         return Mat3(
@@ -316,7 +272,7 @@ public:
         MyBase::mm[8] = 0;
     } // trivial
 
-    /// Set "this" matrix to identity
+    /// Set this matrix to identity
     void setIdentity()
     {
         MyBase::mm[0] = 1;
@@ -341,7 +297,7 @@ public:
         return *this;
     } // opEqualToTest
 
-    /// Test if "this" is equivalent to m with tolerance of eps value
+    /// Return @c true if this matrix is equivalent to @a m within a tolerance of @a eps.
     bool eq(const Mat3 &m, T eps=1.0e-8) const
     {
         return (isApproxEqual(MyBase::mm[0],m.mm[0],eps) &&
@@ -370,7 +326,7 @@ public:
     //     return m*scalar;
     // }
 
-    /// @brief Returns m, where \f$m_{i,j} *= scalar\f$ for \f$i, j \in [0, 2]\f$
+    /// Multiply each element of this matrix by @a scalar.
     template <typename S>
     const Mat3<T>& operator*=(S scalar)
     {
@@ -386,7 +342,7 @@ public:
         return *this;
     }
 
-    /// @brief Returns m0, where \f$m0_{i,j} += m1_{i,j}\f$ for \f$i, j \in [0, 2]\f$
+    /// Add each element of the given matrix to the corresponding element of this matrix.
     template <typename S>
     const Mat3<T> &operator+=(const Mat3<S> &m1)
     {
@@ -404,7 +360,7 @@ public:
         return *this;
     }
 
-    /// @brief Returns m0, where \f$m0_{i,j} -= m1_{i,j}\f$ for \f$i, j \in [0, 2]\f$
+    /// Subtract each element of the given matrix from the corresponding element of this matrix.
     template <typename S>
     const Mat3<T> &operator-=(const Mat3<S> &m1)
     {
@@ -422,7 +378,7 @@ public:
         return *this;
     }
 
-    /// @brief Returns m0, where \f$m0_{i,j} *= m1_{i,j}\f$ for \f$i, j \in [0, 2]\f$
+    /// Multiply this matrix by the given matrix.
     template <typename S>
     const Mat3<T> &operator*=(const Mat3<S> &m1)
     {
@@ -463,7 +419,7 @@ public:
         return *this;
     }
 
-    /// @brief Return the cofactor matrix of "this"
+    /// @brief Return the cofactor matrix of this matrix.
     Mat3 cofactor() const
     {
         return Mat3<T>(
@@ -478,7 +434,7 @@ public:
           MyBase::mm[0] * MyBase::mm[4] - MyBase::mm[1] * MyBase::mm[3]);
     }
 
-    /// returns adjoint of "this", i.e. the transpose of the cofactor of "this" 
+    /// Return the adjoint of this matrix, i.e., the transpose of its cofactor.
     Mat3 adjoint() const
     {
         return Mat3<T>(
@@ -491,9 +447,9 @@ public:
           MyBase::mm[3] * MyBase::mm[7] - MyBase::mm[4] * MyBase::mm[6],
           MyBase::mm[1] * MyBase::mm[6] - MyBase::mm[0] * MyBase::mm[7],
           MyBase::mm[0] * MyBase::mm[4] - MyBase::mm[1] * MyBase::mm[3]);
-        
+
     } // adjointTest
-    
+
     /// returns transpose of this
     Mat3 transpose() const
     {
@@ -512,7 +468,7 @@ public:
 
         const T det = inv.mm[0]*MyBase::mm[0] + inv.mm[1]*MyBase::mm[3] + inv.mm[2]*MyBase::mm[6];
 
-        // If the determinant is 0, m was singular and "this" will contain junk.
+        // If the determinant is 0, m was singular and the result will be invalid.
         if (isApproxEqual(det,T(0.0),tolerance)) {
             OPENVDB_THROW(ArithmeticError, "Inversion of singular 3x3 matrix");
         }
@@ -543,7 +499,7 @@ public:
         return snapMatBasis(*this, axis, direction);
     }
 
-    /// Return the transformed vector by "this" matrix.
+    /// Return the transformed vector by this matrix.
     /// This function is equivalent to post-multiplying the matrix.
     template<typename T0>
     Vec3<T0> transform(const Vec3<T0> &v) const
@@ -551,7 +507,7 @@ public:
         return static_cast< Vec3<T0> >(v * *this);
     } // xformVectorTest
 
-    /// Return the transformed vector by transpose of "this" matrix.
+    /// Return the transformed vector by transpose of this matrix.
     /// This function is equivalent to pre-multiplying the matrix.
     template<typename T0>
     Vec3<T0> pretransform(const Vec3<T0> &v) const
@@ -560,8 +516,8 @@ public:
     } // xformTVectorTest
 
 
-    /// Treats diag as a diagonal matrix and returns the
-    /// multiplication of "this" with diag (from the right).
+    /// @brief Treat @a diag as a diagonal matrix and return the product
+    /// of this matrix with @a diag (from the right).
     Mat3 timesDiagonal(const Vec3<T>& diag) const
     {
         Mat3 ret(*this);
@@ -577,23 +533,9 @@ public:
         ret.mm[8] *= diag(2);
         return ret;
     }
-
-private:
-    static const Mat3<T> sIdentity;
-    static const Mat3<T> sZero;
 }; // class Mat3
 
 
-template <typename T>
-const Mat3<T> Mat3<T>::sIdentity = Mat3<T>(1, 0, 0,
-                                           0, 1, 0,
-                                           0, 0, 1);
-
-template <typename T>
-const Mat3<T> Mat3<T>::sZero = Mat3<T>(0, 0, 0,
-                                       0, 0, 0,
-                                       0, 0, 0);
-
 /// @relates Mat3
 /// @brief Equality operator, does exact floating point comparisons
 template <typename T0, typename T1>
@@ -614,13 +556,13 @@ template <typename T0, typename T1>
 bool operator!=(const Mat3<T0> &m0, const Mat3<T1> &m1) { return !(m0 == m1); }
 
 /// @relates Mat3
-/// @brief Returns M, where \f$M_{i,j} = m_{i,j} * scalar\f$ for \f$i, j \in [0, 2]\f$
+/// @brief Multiply each element of the given matrix by @a scalar and return the result.
 template <typename S, typename T>
 Mat3<typename promote<S, T>::type> operator*(S scalar, const Mat3<T> &m)
 { return m*scalar; }
 
 /// @relates Mat3
-/// @brief Returns M, where \f$M_{i,j} = m_{i,j} * scalar\f$ for \f$i, j \in [0, 2]\f$
+/// @brief Multiply each element of the given matrix by @a scalar and return the result.
 template <typename S, typename T>
 Mat3<typename promote<S, T>::type> operator*(const Mat3<T> &m, S scalar)
 {
@@ -630,7 +572,7 @@ Mat3<typename promote<S, T>::type> operator*(const Mat3<T> &m, S scalar)
 }
 
 /// @relates Mat3
-/// @brief Returns M, where  \f$M_{i,j} = m0_{i,j} + m1_{i,j}\f$ for \f$i, j \in [0, 2]\f$
+/// @brief Add corresponding elements of @a m0 and @a m1 and return the result.
 template <typename T0, typename T1>
 Mat3<typename promote<T0, T1>::type> operator+(const Mat3<T0> &m0, const Mat3<T1> &m1)
 {
@@ -640,7 +582,7 @@ Mat3<typename promote<T0, T1>::type> operator+(const Mat3<T0> &m0, const Mat3<T1
 }
 
 /// @relates Mat3
-/// @brief Returns M, where  \f$M_{i,j} = m0_{i,j} - m1_{i,j}\f$ for \f$i, j \in [0, 2]\f$
+/// @brief Subtract corresponding elements of @a m0 and @a m1 and return the result.
 template <typename T0, typename T1>
 Mat3<typename promote<T0, T1>::type> operator-(const Mat3<T0> &m0, const Mat3<T1> &m1)
 {
@@ -650,10 +592,7 @@ Mat3<typename promote<T0, T1>::type> operator-(const Mat3<T0> &m0, const Mat3<T1
 }
 
 
-/// @brief Matrix multiplication.
-///
-/// Returns M, where
-///     \f$M_{ij} = \sum_{n=0}^2\left(m0_{nj} + m1_{in}\right)\f$ for \f$i, j \in [0, 2]\f$
+/// @brief Multiply @a m0 by @a m1 and return the resulting matrix.
 template <typename T0, typename T1>
 Mat3<typename promote<T0, T1>::type>operator*(const Mat3<T0> &m0, const Mat3<T1> &m1)
 {
@@ -663,7 +602,7 @@ Mat3<typename promote<T0, T1>::type>operator*(const Mat3<T0> &m0, const Mat3<T1>
 }
 
 /// @relates Mat3
-/// @brief Returns v, where \f$v_{i} = \sum_{n=0}^2 m_{i,n} * v_n\f$ for \f$i \in [0, 2]\f$
+/// @brief Multiply @a _m by @a _v and return the resulting vector.
 template<typename T, typename MT>
 inline Vec3<typename promote<T, MT>::type>
 operator*(const Mat3<MT> &_m, const Vec3<T> &_v)
@@ -676,7 +615,7 @@ operator*(const Mat3<MT> &_m, const Vec3<T> &_v)
 }
 
 /// @relates Mat3
-/// @brief Returns v, where \f$v_{i} = \sum_{n=0}^2 m_{n,i} * v_n\f$ for \f$i \in [0, 2]\f$
+/// @brief Multiply @a _v by @a _m and return the resulting vector.
 template<typename T, typename MT>
 inline Vec3<typename promote<T, MT>::type>
 operator*(const Vec3<T> &_v, const Mat3<MT> &_m)
@@ -689,7 +628,7 @@ operator*(const Vec3<T> &_v, const Mat3<MT> &_m)
 }
 
 /// @relates Mat3
-/// @brief Returns v, where \f$v_{i} = \sum_{n=0}^2 m_{i,n} * v_n\f$ for \f$i \in [0, 2]\f$
+/// @brief Multiply @a _v by @a _m and replace @a _v with the resulting vector.
 template<typename T, typename MT>
 inline Vec3<T> &operator *= (Vec3<T> &_v, const Mat3<MT> &_m)
 {
@@ -703,20 +642,11 @@ inline Vec3<T> &operator *= (Vec3<T> &_v, const Mat3<MT> &_m)
 template <typename T>
 Mat3<T> outerProduct(const Vec3<T>& v1, const Vec3<T>& v2)
 {
-    return Mat3<T>(v1[0]*v2[0], v1[0]*v2[1], v1[0]*v2[2], 
-                   v1[1]*v2[0], v1[1]*v2[1], v1[1]*v2[2], 
+    return Mat3<T>(v1[0]*v2[0], v1[0]*v2[1], v1[0]*v2[2],
+                   v1[1]*v2[0], v1[1]*v2[1], v1[1]*v2[2],
                    v1[2]*v2[0], v1[2]*v2[1], v1[2]*v2[2]);
 }// outerProduct
 
-typedef Mat3<float>  Mat3s;
-typedef Mat3<double> Mat3d;
-
-#if DWREAL_IS_DOUBLE == 1
-typedef Mat3d    Mat3f;
-#else
-typedef Mat3s    Mat3f;
-#endif // DWREAL_IS_DOUBLE
-
 
 /// Interpolate the rotation between m1 and m2 using Mat::powSolve.
 /// Unlike slerp, translation is not treated independently.
@@ -731,68 +661,71 @@ Mat3<T> powLerp(const Mat3<T0> &m1, const Mat3<T0> &m2, T t)
 }
 
 
-namespace {
-    template<typename T>
-    void pivot(int i, int j, Mat3<T>& S, Vec3<T>& D, Mat3<T>& Q)
-    {
-        const int& n = Mat3<T>::size;  // should be 3
-        T temp;
-        /// scratch variables used in pivoting
-        double cotan_of_2_theta;
-        double tan_of_theta;
-        double cosin_of_theta;
-        double sin_of_theta;
-        double z;
+namespace mat3_internal {
 
-        double Sij = S(i,j);
+template<typename T>
+inline void
+pivot(int i, int j, Mat3<T>& S, Vec3<T>& D, Mat3<T>& Q)
+{
+    const int& n = Mat3<T>::size;  // should be 3
+    T temp;
+    /// scratch variables used in pivoting
+    double cotan_of_2_theta;
+    double tan_of_theta;
+    double cosin_of_theta;
+    double sin_of_theta;
+    double z;
 
-        double Sjj_minus_Sii = D[j] - D[i];
+    double Sij = S(i,j);
 
-        if (fabs(Sjj_minus_Sii) * (10*math::Tolerance<T>::value()) > fabs(Sij)) {
-            tan_of_theta = Sij / Sjj_minus_Sii;
+    double Sjj_minus_Sii = D[j] - D[i];
+
+    if (fabs(Sjj_minus_Sii) * (10*math::Tolerance<T>::value()) > fabs(Sij)) {
+        tan_of_theta = Sij / Sjj_minus_Sii;
+    } else {
+        /// pivot on Sij
+        cotan_of_2_theta = 0.5*Sjj_minus_Sii / Sij ;
+
+        if (cotan_of_2_theta < 0.) {
+            tan_of_theta =
+                -1./(sqrt(1. + cotan_of_2_theta*cotan_of_2_theta) - cotan_of_2_theta);
         } else {
-            /// pivot on Sij
-            cotan_of_2_theta = 0.5*Sjj_minus_Sii / Sij ;
-
-            if (cotan_of_2_theta < 0.) {
-                tan_of_theta =
-                    -1./(sqrt(1. + cotan_of_2_theta*cotan_of_2_theta) - cotan_of_2_theta);
-            } else {
-                tan_of_theta =
-                    1./(sqrt(1. + cotan_of_2_theta*cotan_of_2_theta) + cotan_of_2_theta);
-            }
+            tan_of_theta =
+                1./(sqrt(1. + cotan_of_2_theta*cotan_of_2_theta) + cotan_of_2_theta);
         }
-
-        cosin_of_theta = 1./sqrt( 1. + tan_of_theta * tan_of_theta);
-        sin_of_theta = cosin_of_theta * tan_of_theta;
-        z = tan_of_theta * Sij;
-        S(i,j) = 0;
-        D[i] -= z;
-        D[j] += z;
-        for (int k = 0; k < i; ++k) {
-            temp = S(k,i);
-            S(k,i) = cosin_of_theta * temp - sin_of_theta * S(k,j);
-            S(k,j)= sin_of_theta * temp + cosin_of_theta * S(k,j);
-        }
-        for (int k = i+1; k < j; ++k) {
-            temp = S(i,k);
-            S(i,k) = cosin_of_theta * temp - sin_of_theta * S(k,j);
-            S(k,j) = sin_of_theta * temp + cosin_of_theta * S(k,j);
-        }
-        for (int k = j+1; k < n; ++k) {
-            temp = S(i,k);
-            S(i,k) = cosin_of_theta * temp - sin_of_theta * S(j,k);
-            S(j,k) = sin_of_theta * temp + cosin_of_theta * S(j,k);
-        }
-        for (int k = 0; k < n; ++k)
-            {
-                temp = Q(k,i);
-                Q(k,i) = cosin_of_theta * temp - sin_of_theta*Q(k,j);
-                Q(k,j) = sin_of_theta * temp + cosin_of_theta*Q(k,j);
-            }
     }
+
+    cosin_of_theta = 1./sqrt( 1. + tan_of_theta * tan_of_theta);
+    sin_of_theta = cosin_of_theta * tan_of_theta;
+    z = tan_of_theta * Sij;
+    S(i,j) = 0;
+    D[i] -= z;
+    D[j] += z;
+    for (int k = 0; k < i; ++k) {
+        temp = S(k,i);
+        S(k,i) = cosin_of_theta * temp - sin_of_theta * S(k,j);
+        S(k,j)= sin_of_theta * temp + cosin_of_theta * S(k,j);
+    }
+    for (int k = i+1; k < j; ++k) {
+        temp = S(i,k);
+        S(i,k) = cosin_of_theta * temp - sin_of_theta * S(k,j);
+        S(k,j) = sin_of_theta * temp + cosin_of_theta * S(k,j);
+    }
+    for (int k = j+1; k < n; ++k) {
+        temp = S(i,k);
+        S(i,k) = cosin_of_theta * temp - sin_of_theta * S(j,k);
+        S(j,k) = sin_of_theta * temp + cosin_of_theta * S(j,k);
+    }
+    for (int k = 0; k < n; ++k)
+        {
+            temp = Q(k,i);
+            Q(k,i) = cosin_of_theta * temp - sin_of_theta*Q(k,j);
+            Q(k,j) = sin_of_theta * temp + cosin_of_theta*Q(k,j);
+        }
 }
 
+} // namespace mat3_internal
+
 
 /// @brief Use Jacobi iterations to decompose a symmetric 3x3 matrix
 /// (diagonalize and compute eigenvectors)
@@ -800,7 +733,8 @@ namespace {
 /// Joachim Kopp.  arXiv.org preprint: physics/0610206
 /// with the addition of largest pivot
 template<typename T>
-bool diagonalizeSymmetricMatrix(const Mat3<T>& input, Mat3<T>& Q, Vec3<T>& D,
+inline bool
+diagonalizeSymmetricMatrix(const Mat3<T>& input, Mat3<T>& Q, Vec3<T>& D,
     unsigned int MAX_ITERATIONS=250)
 {
     /// use Givens rotation matrix to eliminate off-diagonal entries.
@@ -850,18 +784,66 @@ bool diagonalizeSymmetricMatrix(const Mat3<T>& input, Mat3<T>& Q, Vec3<T>& D,
                 }
             }
         }
-        pivot(ip, jp, S, D, Q);
+        mat3_internal::pivot(ip, jp, S, D, Q);
     } while (iterations < MAX_ITERATIONS);
 
     return false;
 }
 
+template<typename T>
+inline Mat3<T>
+Abs(const Mat3<T>& m)
+{
+    Mat3<T> out;
+    const T* ip = m.asPointer();
+    T* op = out.asPointer();
+    for (unsigned i = 0; i < 9; ++i, ++op, ++ip) *op = math::Abs(*ip);
+    return out;
+}
+
+template<typename Type1, typename Type2>
+inline Mat3<Type1>
+cwiseAdd(const Mat3<Type1>& m, const Type2 s)
+{
+    Mat3<Type1> out;
+    const Type1* ip = m.asPointer();
+    Type1* op = out.asPointer();
+    for (unsigned i = 0; i < 9; ++i, ++op, ++ip) {
+        OPENVDB_NO_TYPE_CONVERSION_WARNING_BEGIN
+        *op = *ip + s;
+        OPENVDB_NO_TYPE_CONVERSION_WARNING_END
+    }
+    return out;
+}
+
+template<typename T>
+inline bool
+cwiseLessThan(const Mat3<T>& m0, const Mat3<T>& m1)
+{
+    return cwiseLessThan<3, T>(m0, m1);
+}
+
+template<typename T>
+inline bool
+cwiseGreaterThan(const Mat3<T>& m0, const Mat3<T>& m1)
+{
+    return cwiseGreaterThan<3, T>(m0, m1);
+}
+
+using Mat3s = Mat3<float>;
+using Mat3d = Mat3<double>;
+using Mat3f = Mat3d;
+
+OPENVDB_IS_POD(Mat3s)
+OPENVDB_IS_POD(Mat3d)
+
 } // namespace math
+
+
+template<> inline math::Mat3s zeroVal<math::Mat3s>() { return math::Mat3s::zero(); }
+template<> inline math::Mat3d zeroVal<math::Mat3d>() { return math::Mat3d::zero(); }
+
 } // namespace OPENVDB_VERSION_NAME
 } // namespace openvdb
 
 #endif // OPENVDB_MATH_MAT3_H_HAS_BEEN_INCLUDED
-
-// Copyright (c) 2012-2016 DreamWorks Animation LLC
-// All rights reserved. This software is distributed under the
-// Mozilla Public License 2.0 ( http://www.mozilla.org/MPL/2.0/ )
